#
# xmlBlaster core socket functionality
#

import TcpClient

class XBsession:
    def __init__(self,host='localhost',port=7607):
        self.cli = TcpClient.skt(host,port)
        lskt,lport = self.cli.localhostport()
        self.locskt = 'socket:'+lskt+':'+str(lport)
        self.myID = ''
        self.queue = []
        self.conQos = '''<qos>
  <securityService type='htpasswd' version='1.0'>
    <user>%s</user>
    <passwd>%s</passwd>
  </securityService>
  <address type='SOCKET'>
    %s
  </address>
</qos>'''
        self.conQos1way = '''<qos>
  <securityService type='htpasswd' version='1.0'>
    <user>%s</user>
    <passwd>%s</passwd>
  </securityService>
  <queue relating='callback' maxEntries='100' maxEntriesCache='100'>
    <callback type='SOCKET' oneway='true' />
      %s
  </queue>
</qos>'''
        self.clrQos = '''<qos>
  <deleteSubjectQueue>true</deleteSubjectQueue>
  <clearSessions>false</clearSessions>
</qos>'''
        self.callbackQos = '''<qos>
  <queue relating='callback' maxEntries='100' maxEntriesCache='100'>
    <callback type='SOCKET' />
      %s
  </queue>
</qos>'''
        self.ptpCallbackQos = '''<qos>
  <securityService type='htpasswd' version='1.0'>
    <user>%s</user>
    <passwd>%s</passwd>
  </securityService>
  <ptp>true</ptp>
  <queue relating='callback' maxEntries='100' maxEntriesCache='100'>
    <callback type='SOCKET'>
      %s
    </callback>
  </queue>
  <address type='SOCKET'>
    %s
  </address>
</qos>'''
        self.ptpQos = '''<qos>
  <destination queryType='EXACT'>
    %s
  </destination>
</qos>'''

    def encapsulate(self):
        return '%10d%s' % (len(self.payload)+10,self.payload)

    def setflag(self,type='I'):
        cs = chr(0)
        cmprn = chr(0)
        null = chr(0)
        proto = '1'
        return ''.join([cs,cmprn,type,null,null,proto])

    def userData(self,qos,key,content):
        return chr(0).join([qos,key,str(len(content)),content])

    def topicoid(self,topic):
        return "<key oid='"+topic+"'/>"

    def msg(self,rqid,mthd,sess,qos,oid,data,flag='I'):
        self.payload = ''.join([self.setflag(flag),rqid,chr(0),\
            mthd,chr(0),sess,chr(0),'',chr(0),\
            self.userData(qos,oid,data),''])
        return self.encapsulate()

    def ping(self):
        self.cli.write(self.msg(self.myID,'ping','','','',''))
        self.cli.read(4096)

    def connect(self,user,pwd):
        self.cli.write(self.msg(self.myID,'connect','',\
            self.conQos % (user, pwd, self.locskt), '', ''))
        reply = self.cli.read(4096)
        self.sid = reply.split("sessionId='",1)[1].split("'")[0]

    def connectOneway(self,user,pwd):
        self.cli.write(self.msg(self.myID,'connect','',\
            self.conQos1way % (user, pwd, self.locskt), '', ''))
        reply = self.cli.read(4096)
        self.sid = reply.split("sessionId='",1)[1].split("'")[0]

    def publish(self,topic,info):
        self.cli.write(self.msg(self.myID,'publish',self.sid,\
            '<qos/>',self.topicoid(topic),info))
        while 1:
            reply = self.cli.read(4096)
            if 'publish' in reply:
                break
            self.queue.append(reply)

    def publishOneway(self,topic,info):
        self.cli.write(self.msg(self.myID,'publishOneway',self.sid,\
            '<qos/>',self.topicoid(topic),info))

    def get(self,topic):
        self.cli.write(self.msg(self.myID,'get',self.sid,\
            '<qos/>',self.topicoid(topic),''))
        self.cli.read(4096)

    def connectcb(self,user,pwd):
        self.cli.write(self.msg(self.myID,'connect','',\
            self.ptpCallbackQos % (user, pwd, self.locskt, self.locskt), '', ''))
        reply = self.cli.read(4096)
        self.sid = reply.split("sessionId='",1)[1].split("'")[0]

    def subscribe(self,topic):
        self.cli.write(self.msg(self.myID,'subscribe',self.sid,\
            '<qos/>', self.topicoid(topic), ''))
        while 1:
            reply = self.cli.read(4096)
            if 'subscribe' in reply:
                break
            self.queue.append(reply)

    def localupdate(self, ack=0):
        while self.queue:
            reply = self.queue.pop(0)
            if not reply:
                return   # lost connection
            if 'update' in reply:
                if ack:
                    self._updateAck(reply)
                return
        while 1:
            reply = self.cli.read(4096)
            if not reply:
                return   # lost connection
            if 'update' in reply:
                if ack:
                    self._updateAck(reply)
                return

    def _updateAck(self,reply):
        ureqid = reply.split(chr(0)+'update',1)[0][16:]
        usid = reply.split('update'+chr(0),1)[1]\
                .split(chr(0),1)[0]
        self.cli.write(self.msg(ureqid,'update',usid,\
            '<qos/>', '', '', 'R'))

    def decodequery(self,query):
        try:
            count = int(query[:10].strip())
            if count != len(query):
                raise
        except:
            print 'bad query protocol received'
            return ''
        flags = query[10:16]
        if not((flags[2] == 'I') and (flags[5] == '1')):
            print 'bad query flags received'
            return ''
        try:
            rqID, mthdN, ssess, luzip, udata = query[16:].split(chr(0),4)
        except:
            print 'bad query format received'
            return ''
        try:
            qos, key, lenzdata = udata.split(chr(0),2)
        except:
            print 'bad query userdata format'
            return ''
        try:
            lnd, data = lenzdata.split(chr(0),1)
            if int(lnd) != len(data):
                raise
        except:
            print 'bad data length/data format'
            return ''
        return data

    def remoteupdate(self):
        while self.queue:
            query = self.queue.pop(0)
            if not query:
                return ''  # lost connection
            if 'update' in query:
                self._remUpdateAck(query)
                return self.decodequery(query)
        while 1:
            query = self.cli.read(4096)
            if not query:
                return ''  # lost connection
            if 'update' in query:
                self._remUpdateAck(query)
                return self.decodequery(query)

    def _remUpdateAck(self,query):
        ureqid = query.split(chr(0)+'update',1)[0][16:]
        usid = query.split('update'+chr(0),1)[1]\
                .split(chr(0),1)[0]
        self.cli.write(self.msg(ureqid,'update',usid,\
            '<qos/>', '', '', 'R'))

    def publishptp(self,topic,info,whoto):
        self.cli.write(self.msg(self.myID,'publish',self.sid,\
            self.ptpQos % whoto,self.topicoid(topic),info))
        self.cli.read(4096)

    def disconnect(self):
        self.cli.write(self.msg(self.myID,'disconnect',\
            self.sid,self.clrQos,'',''))
        self.cli.read(4096)
